import json

import flask

from ._internal import client_lib, comm, const, errors, execution, function, net


def serve(
    app: flask.Flask,
    client: client_lib.Inngest,
    functions: list[function.Function],
    *,
    base_url: str | None = None,
    signing_key: str | None = None,
) -> None:
    handler = comm.CommHandler(
        base_url=base_url or client.base_url,
        client=client,
        framework=const.Framework.FLASK,
        functions=functions,
        logger=app.logger,
        signing_key=signing_key,
    )

    async_mode = any(
        function.is_handler_async or function.is_on_failure_handler_async
        for function in functions
    )
    if async_mode:
        _create_handler_async(app, client, handler)
    else:
        _create_handler_sync(app, client, handler)


def _create_handler_async(
    app: flask.Flask,
    client: client_lib.Inngest,
    handler: comm.CommHandler,
) -> None:
    @app.route("/api/inngest", methods=["GET", "POST", "PUT"])
    async def inngest_api() -> flask.Response | str:
        headers = net.normalize_headers(dict(flask.request.headers.items()))
        is_from_dev_server = (
            headers.get(const.HeaderKey.SERVER_KIND.value)
            == const.ServerKind.DEV_SERVER.value
        )

        if flask.request.method == "GET":
            return _to_response(handler.inspect(is_from_dev_server))

        if flask.request.method == "POST":
            fn_id = flask.request.args.get("fnId")
            if fn_id is None:
                raise errors.MissingParam("fnId")

            return _to_response(
                await handler.call_function(
                    call=execution.Call.from_dict(
                        json.loads(flask.request.data)
                    ),
                    fn_id=fn_id,
                    req_sig=net.RequestSignature(
                        body=flask.request.data,
                        headers=headers,
                        is_production=client.is_production,
                    ),
                )
            )

        if flask.request.method == "PUT":
            return _to_response(
                await handler.register(
                    app_url=flask.request.url,
                    is_from_dev_server=is_from_dev_server,
                )
            )

        # Should be unreachable
        return ""


def _create_handler_sync(
    app: flask.Flask,
    client: client_lib.Inngest,
    handler: comm.CommHandler,
) -> None:
    @app.route("/api/inngest", methods=["GET", "POST", "PUT"])
    def inngest_api() -> flask.Response | str:
        headers = net.normalize_headers(dict(flask.request.headers.items()))
        is_from_dev_server = (
            headers.get(const.HeaderKey.SERVER_KIND.value)
            == const.ServerKind.DEV_SERVER.value
        )

        if flask.request.method == "GET":
            return _to_response(handler.inspect(is_from_dev_server))

        if flask.request.method == "POST":
            fn_id = flask.request.args.get("fnId")
            if fn_id is None:
                raise errors.MissingParam("fnId")

            return _to_response(
                handler.call_function_sync(
                    call=execution.Call.from_dict(
                        json.loads(flask.request.data)
                    ),
                    fn_id=fn_id,
                    req_sig=net.RequestSignature(
                        body=flask.request.data,
                        headers=headers,
                        is_production=client.is_production,
                    ),
                )
            )

        if flask.request.method == "PUT":
            return _to_response(
                handler.register_sync(
                    app_url=flask.request.url,
                    is_from_dev_server=is_from_dev_server,
                )
            )

        # Should be unreachable
        return ""


def _to_response(comm_res: comm.CommResponse) -> flask.Response:
    res = flask.make_response()

    for k, v in comm_res.headers.items():
        res.headers.add_header(k, v)

    res.set_data(json.dumps(comm_res.body))
    res.status_code = comm_res.status_code
    return res



from __future__ import annotations

import typing

ErrT = typing.TypeVar("ErrT")
OkT = typing.TypeVar("OkT")


class Err(typing.Generic[ErrT]):
    __match_args__ = ("err_value",)
    __slots__ = ("_value",)

    def __init__(self, value: ErrT):
        self._value = value

    @property
    def err_value(self) -> ErrT:
        return self._value

    def is_err(self) -> typing.Literal[True]:
        return True

    def is_ok(self) -> typing.Literal[False]:
        return False


class Ok(typing.Generic[OkT]):
    __match_args__ = ("ok_value",)
    __slots__ = ("_value",)

    def __init__(self, value: OkT):
        self._value = value

    @property
    def ok_value(self) -> OkT:
        return self._value

    def apply_to_ok(self, fn: typing.Callable[[OkT], OkT]) -> Ok[OkT]:
        """
        Apply a transform function to the Ok value and return a new instance of
        Ok.
        """

        return Ok(fn(self.ok_value))

    def is_err(self) -> typing.Literal[False]:
        return False

    def is_ok(self) -> typing.Literal[True]:
        return True


Result: typing.TypeAlias = Ok[OkT] | Err[ErrT]


def is_err(result: Result[OkT, ErrT]) -> typing.TypeGuard[Err[ErrT]]:
    return result.is_err()


def is_ok(result: Result[OkT, ErrT]) -> typing.TypeGuard[Ok[OkT]]:
    return result.is_ok()



import datetime
import inspect
import typing

from inngest._internal import (
    client_lib,
    event_lib,
    execution,
    result,
    transforms,
    types,
)

from . import base


class Step(base.StepBase):
    def __init__(
        self,
        client: client_lib.Inngest,
        memos: dict[str, object],
        step_id_counter: base.StepIDCounter,
    ) -> None:
        self._client = client
        self._memos = memos
        self._step_id_counter = step_id_counter

    @typing.overload
    async def run(
        self,
        step_id: str,
        handler: typing.Callable[[], typing.Awaitable[types.SerializableT]],
    ) -> types.SerializableT:
        ...

    @typing.overload
    async def run(
        self,
        step_id: str,
        handler: typing.Callable[[], types.SerializableT],
    ) -> types.SerializableT:
        ...

    async def run(
        self,
        step_id: str,
        handler: typing.Callable[[], typing.Awaitable[types.SerializableT]]
        | typing.Callable[[], types.SerializableT],
    ) -> types.SerializableT:
        """
        Run logic that should be retried on error and memoized after success.

        Args:
            step_id: Unique step ID within the function. If the same step ID is
                encountered multiple times then it'll get an index suffix.
            handler: The logic to run. Can be async or sync.
        """

        hashed_id = self._get_hashed_id(step_id)

        memo = self._get_memo(hashed_id)
        if memo is not types.EmptySentinel:
            return memo  # type: ignore

        if inspect.iscoroutinefunction(handler):
            output = await handler()
        else:
            output = handler()

        # Check whether output is serializable
        match transforms.dump_json(output):
            case result.Ok(_):
                pass
            case result.Err(err):
                raise err

        raise base.Interrupt(
            hashed_id=hashed_id,
            data=output,
            display_name=step_id,
            op=execution.Opcode.STEP,
            name=step_id,
        )

    async def send_event(
        self,
        step_id: str,
        events: event_lib.Event | list[event_lib.Event],
    ) -> list[str]:
        """
        Send an event or list of events.
        """

        async def fn() -> list[str]:
            return await self._client.send(events)

        return await self.run(step_id, fn)

    async def sleep(
        self,
        step_id: str,
        duration: int | datetime.timedelta,
    ) -> None:
        """
        Sleep for a duration.

        Args:
            duration: The number of milliseconds to sleep.
        """

        if isinstance(duration, int):
            until = datetime.datetime.utcnow() + datetime.timedelta(
                milliseconds=duration
            )
        else:
            until = datetime.datetime.utcnow() + duration

        return await self.sleep_until(step_id, until)

    async def sleep_until(
        self,
        step_id: str,
        until: datetime.datetime,
    ) -> None:
        """
        Sleep until a specific time.
        """

        hashed_id = self._get_hashed_id(step_id)

        memo = self._get_memo(hashed_id)
        if memo is not types.EmptySentinel:
            return memo  # type: ignore

        raise base.Interrupt(
            hashed_id=hashed_id,
            display_name=step_id,
            name=transforms.to_iso_utc(until),
            op=execution.Opcode.SLEEP,
        )

    async def wait_for_event(
        self,
        step_id: str,
        *,
        event: str,
        if_exp: str | None = None,
        timeout: int | datetime.timedelta,
    ) -> event_lib.Event | None:
        """
        Wait for an event to be sent.

        Args:
            event: Event name.
            if_exp: An expression to filter events.
            timeout: The maximum number of milliseconds to wait for the event.
        """

        hashed_id = self._get_hashed_id(step_id)

        memo = self._get_memo(hashed_id)
        if memo is not types.EmptySentinel:
            if memo is None:
                # Timeout
                return None

            # Fulfilled by an event
            return event_lib.Event.model_validate(memo)

        match transforms.to_duration_str(timeout):
            case result.Ok(timeout_str):
                pass
            case result.Err(err):
                raise err

        match base.WaitForEventOpts(
            if_exp=if_exp,
            timeout=timeout_str,
        ).to_dict():
            case result.Ok(opts):
                pass
            case result.Err(err):
                raise err

        raise base.Interrupt(
            hashed_id=hashed_id,
            display_name=step_id,
            name=event,
            op=execution.Opcode.WAIT_FOR_EVENT,
            opts=opts,
        )



import json
import typing

import tornado.web

from inngest._internal import (
    client_lib,
    comm,
    const,
    errors,
    execution,
    function,
    net,
)


def serve(
    app: tornado.web.Application,
    client: client_lib.Inngest,
    functions: list[function.Function],
    *,
    base_url: str | None = None,
    signing_key: str | None = None,
) -> None:
    handler = comm.CommHandler(
        base_url=base_url or client.base_url,
        client=client,
        framework=const.Framework.TORNADO,
        functions=functions,
        logger=client.logger,
        signing_key=signing_key,
    )

    class InngestHandler(tornado.web.RequestHandler):
        def data_received(self, chunk: bytes) -> typing.Awaitable[None] | None:
            return None

        def get(self) -> None:
            headers = net.normalize_headers(dict(self.request.headers.items()))
            is_from_dev_server = (
                headers.get(const.HeaderKey.SERVER_KIND.value)
                == const.ServerKind.DEV_SERVER.value
            )

            comm_res = handler.inspect(is_from_dev_server)
            self.write(json.dumps(comm_res.body))
            for k, v in comm_res.headers.items():
                self.add_header(k, v)
            self.set_status(comm_res.status_code)

        def post(self) -> None:
            fn_id: str | None
            raw_fn_id = self.request.query_arguments.get("fnId")
            if raw_fn_id is None or len(raw_fn_id) == 0:
                raise errors.MissingParam("fnId")
            fn_id = raw_fn_id[0].decode("utf-8")

            headers = net.normalize_headers(dict(self.request.headers.items()))

            comm_res = handler.call_function_sync(
                call=execution.Call.from_dict(json.loads(self.request.body)),
                fn_id=fn_id,
                req_sig=net.RequestSignature(
                    body=self.request.body,
                    headers=headers,
                    is_production=client.is_production,
                ),
            )
            self.write(json.dumps(comm_res.body))
            for k, v in comm_res.headers.items():
                self.add_header(k, v)
            self.set_status(comm_res.status_code)

        def put(self) -> None:
            headers = net.normalize_headers(dict(self.request.headers.items()))

            comm_res = handler.register_sync(
                app_url=self.request.full_url(),
                is_from_dev_server=(
                    headers.get(const.HeaderKey.SERVER_KIND.value)
                    == const.ServerKind.DEV_SERVER.value
                ),
            )
            self.write(json.dumps(comm_res.body))
            for k, v in comm_res.headers.items():
                self.add_header(k, v)
            self.set_status(comm_res.status_code)

    app.add_handlers(r".*", [("/api/inngest", InngestHandler)])



from __future__ import annotations

import typing

import pydantic

from . import result

T = typing.TypeVar("T")

EmptySentinel = object()

Serializable = (
    bool
    | float
    | int
    | str
    | dict[typing.Any, typing.Any]
    | list[typing.Any]
    | tuple[typing.Any, ...]
    | None
)
SerializableT = typing.TypeVar("SerializableT", bound=Serializable)


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(strict=True)

    def __init__(  # pylint: disable=no-self-argument
        __pydantic_self__,
        *args: object,
        **kwargs: object,
    ) -> None:
        try:
            super().__init__(*args, **kwargs)
        except pydantic.ValidationError as err:
            raise __pydantic_self__.convert_validation_error(err) from None

    def convert_validation_error(
        self,
        err: pydantic.ValidationError,
    ) -> BaseException:
        """
        Subclasses can override this method to convert Pydantic's
        ValidationError into a different error.
        """

        return err

    @classmethod
    def from_dict(
        cls: typing.Type[BaseModelT],
        raw: dict[str, object],
    ) -> BaseModelT:
        return cls.model_validate(raw)

    def to_dict(self) -> result.Result[dict[str, object], Exception]:
        try:
            return result.Ok(self.model_dump(mode="json"))
        except Exception as err:
            return result.Err(err)


BaseModelT = typing.TypeVar("BaseModelT", bound=BaseModel)



from __future__ import annotations

import logging
import os
import time
import urllib.parse

import httpx

from . import const, env, errors, event_lib, net, result


class Inngest:
    def __init__(
        self,
        *,
        app_id: str,
        base_url: str | None = None,
        event_key: str | None = None,
        is_production: bool | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        self.app_id = app_id
        self.base_url = base_url
        self.is_production = is_production or env.is_prod()
        self.logger = logger or logging.getLogger(__name__)

        if event_key is None:
            if not self.is_production:
                event_key = "NO_EVENT_KEY_SET"
            else:
                event_key = os.getenv(const.EnvKey.EVENT_KEY.value)
        if event_key is None:
            self.logger.error("missing event key")
            raise errors.MissingEventKey()
        self._event_key = event_key

        event_origin = base_url
        if event_origin is None:
            if not self.is_production:
                self.logger.info("Defaulting event origin to Dev Server")
                event_origin = const.DEV_SERVER_ORIGIN
            else:
                event_origin = const.DEFAULT_EVENT_ORIGIN
        self._event_origin = event_origin

    def _build_send_request(
        self,
        events: list[event_lib.Event],
    ) -> result.Result[httpx.Request, Exception]:
        url = urllib.parse.urljoin(self._event_origin, f"/e/{self._event_key}")
        headers = net.create_headers()

        body = []
        for event in events:
            match event.to_dict():
                case result.Ok(d):
                    pass
                case result.Err(err):
                    return result.Err(err)

            if d.get("id") == "":
                del d["id"]
            if d.get("ts") == 0:
                d["ts"] = int(time.time() * 1000)
            body.append(d)

        return result.Ok(
            httpx.Client().build_request(
                "POST",
                url,
                headers=headers,
                json=body,
                timeout=30,
            )
        )

    async def send(
        self,
        events: event_lib.Event | list[event_lib.Event],
    ) -> list[str]:
        if not isinstance(events, list):
            events = [events]

        async with httpx.AsyncClient() as client:
            match self._build_send_request(events):
                case result.Ok(req):
                    ids = _extract_ids((await client.send(req)).json())
                case result.Err(err):
                    raise err
        return ids

    def send_sync(
        self,
        events: event_lib.Event | list[event_lib.Event],
    ) -> list[str]:
        if not isinstance(events, list):
            events = [events]

        with httpx.Client() as client:
            match self._build_send_request(events):
                case result.Ok(req):
                    ids = _extract_ids(client.send(req).json())
                case result.Err(err):
                    raise err
        return ids


def _extract_ids(body: object) -> list[str]:
    if not isinstance(body, dict) or "ids" not in body:
        raise errors.InvalidBody("unexpected response when sending events")

    ids = body["ids"]
    if not isinstance(ids, list):
        raise errors.InvalidBody("unexpected response when sending events")

    return ids



import json
import unittest

import flask
import flask.testing

import inngest
import inngest.flask
from inngest._internal import const

from . import base, cases, dev_server, http_proxy, net

_client = inngest.Inngest(
    app_id="flask",
    base_url=f"http://{net.HOST}:{dev_server.PORT}",
)

_cases = cases.create_cases_sync(_client, "flask")


class TestFlask(unittest.TestCase):
    app: flask.testing.FlaskClient
    dev_server_port: int
    proxy: http_proxy.Proxy

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        app = flask.Flask(__name__)
        app.logger.disabled = True
        inngest.flask.serve(
            app,
            _client,
            [
                case.fn
                for case in _cases
                # Should always be true but mypy doesn't know that
                if isinstance(case.fn, inngest.Function)
            ],
        )
        cls.app = app.test_client()
        cls.proxy = http_proxy.Proxy(cls.on_proxy_request).start()
        base.register(cls.proxy.port)

    @classmethod
    def tearDownClass(cls) -> None:
        super().tearDownClass()
        cls.proxy.stop()

    @classmethod
    def on_proxy_request(
        cls,
        *,
        body: bytes | None,
        headers: dict[str, list[str]],
        method: str,
        path: str,
    ) -> http_proxy.Response:
        res = cls.app.open(
            method=method,
            path=path,
            headers=headers,
            data=body,
        )

        return http_proxy.Response(
            body=res.data,
            headers=dict(res.headers),
            status_code=res.status_code,
        )


for case in _cases:
    test_name = f"test_{case.name}"
    setattr(TestFlask, test_name, case.run_test)


class TestRegistration(unittest.TestCase):
    def test_dev_server_to_prod(self) -> None:
        """
        Ensure that Dev Server cannot initiate a registration request when in
        production mode.
        """

        client = inngest.Inngest(
            app_id="flask_registration",
            event_key="test",
            is_production=True,
        )

        @inngest.create_function(
            fn_id="foo",
            retries=0,
            trigger=inngest.TriggerEvent(event="app/foo"),
        )
        def fn(**_kwargs: object) -> None:
            pass

        app = flask.Flask(__name__)
        inngest.flask.serve(
            app,
            client,
            [fn],
            signing_key="signkey-prod-0486c9",
        )
        flask_client = app.test_client()
        res = flask_client.put(
            "/api/inngest",
            headers={
                const.HeaderKey.SERVER_KIND.value.lower(): const.ServerKind.DEV_SERVER.value,
            },
        )
        assert res.status_code == 400
        body: object = json.loads(res.data)
        assert (
            isinstance(body, dict)
            and body["code"]
            == const.ErrorCode.DISALLOWED_REGISTRATION_INITIATOR.value
        )


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import dataclasses
import hashlib
import inspect
import typing

import pydantic

from inngest._internal import (
    client_lib,
    const,
    errors,
    event_lib,
    execution,
    function_config,
    step_lib,
    transforms,
    types,
)


@dataclasses.dataclass
class _Config:
    # The user-defined function
    main: function_config.FunctionConfig

    # The internal on_failure function
    on_failure: function_config.FunctionConfig | None


@typing.runtime_checkable
class FunctionHandlerAsync(typing.Protocol):
    def __call__(
        self,
        *,
        attempt: int,
        event: event_lib.Event,
        events: list[event_lib.Event],
        run_id: str,
        step: step_lib.Step,
    ) -> typing.Awaitable[types.Serializable]:
        ...


@typing.runtime_checkable
class FunctionHandlerSync(typing.Protocol):
    def __call__(
        self,
        *,
        attempt: int,
        event: event_lib.Event,
        events: list[event_lib.Event],
        run_id: str,
        step: step_lib.StepSync,
    ) -> types.Serializable:
        ...


def _is_function_handler_async(
    value: FunctionHandlerAsync | FunctionHandlerSync,
) -> typing.TypeGuard[FunctionHandlerAsync]:
    return inspect.iscoroutinefunction(value)


def _is_function_handler_sync(
    value: FunctionHandlerAsync | FunctionHandlerSync,
) -> typing.TypeGuard[FunctionHandlerSync]:
    return not inspect.iscoroutinefunction(value)


class FunctionOpts(types.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    batch_events: function_config.Batch | None = None
    cancel: list[function_config.Cancel] | None = None
    debounce: function_config.Debounce | None = None
    id: str
    name: str | None = None
    on_failure: FunctionHandlerAsync | FunctionHandlerSync | None = None
    rate_limit: function_config.RateLimit | None = None
    retries: int | None = None
    throttle: function_config.Throttle | None = None

    def convert_validation_error(
        self,
        err: pydantic.ValidationError,
    ) -> BaseException:
        return errors.InvalidConfig.from_validation_error(err)


def create_function(
    *,
    batch_events: function_config.Batch | None = None,
    cancel: list[function_config.Cancel] | None = None,
    debounce: function_config.Debounce | None = None,
    fn_id: str,
    name: str | None = None,
    on_failure: FunctionHandlerAsync | FunctionHandlerSync | None = None,
    rate_limit: function_config.RateLimit | None = None,
    retries: int | None = None,
    throttle: function_config.Throttle | None = None,
    trigger: function_config.TriggerCron | function_config.TriggerEvent,
) -> typing.Callable[[FunctionHandlerAsync | FunctionHandlerSync], Function]:
    def decorator(func: FunctionHandlerAsync | FunctionHandlerSync) -> Function:
        return Function(
            FunctionOpts(
                batch_events=batch_events,
                cancel=cancel,
                debounce=debounce,
                id=fn_id,
                name=name,
                on_failure=on_failure,
                rate_limit=rate_limit,
                retries=retries,
                throttle=throttle,
            ),
            trigger,
            func,
        )

    return decorator


class Function:
    _handler: FunctionHandlerAsync | FunctionHandlerSync
    _on_failure_fn_id: str | None = None
    _opts: FunctionOpts
    _trigger: function_config.TriggerCron | function_config.TriggerEvent

    @property
    def id(self) -> str:
        return self._opts.id

    @property
    def is_handler_async(self) -> bool:
        """
        Whether the main handler is async.
        """
        return _is_function_handler_async(self._handler)

    @property
    def is_on_failure_handler_async(self) -> bool | None:
        """
        Whether the on_failure handler is async. Returns None if there isn't an
        on_failure handler.
        """

        if self._opts.on_failure is None:
            return None
        return _is_function_handler_async(self._opts.on_failure)

    @property
    def on_failure_fn_id(self) -> str | None:
        return self._on_failure_fn_id

    def __init__(
        self,
        opts: FunctionOpts,
        trigger: function_config.TriggerCron | function_config.TriggerEvent,
        handler: FunctionHandlerAsync | FunctionHandlerSync,
    ) -> None:
        self._handler = handler
        self._opts = opts
        self._trigger = trigger

        if opts.on_failure is not None:
            # Create a random suffix to avoid collisions with the main
            # function's ID.
            suffix = hashlib.sha1(opts.id.encode("utf-8")).hexdigest()[:8]

            self._on_failure_fn_id = f"{opts.id}-{suffix}"

    async def call(
        self,
        call: execution.Call,
        client: client_lib.Inngest,
        fn_id: str,
    ) -> list[
        execution.CallResponse
    ] | types.Serializable | execution.CallError:
        try:
            handler: FunctionHandlerAsync | FunctionHandlerSync
            if self.id == fn_id:
                handler = self._handler
            elif self.on_failure_fn_id == fn_id:
                if self._opts.on_failure is None:
                    return execution.CallError.from_error(
                        errors.MissingFunction("on_failure not defined")
                    )
                handler = self._opts.on_failure
            else:
                return execution.CallError.from_error(
                    errors.MissingFunction("function ID mismatch")
                )

            # Determine whether the handler is async (i.e. if we need to await
            # it). Sync functions are OK in async contexts, so it's OK if the
            # handler is sync.
            if _is_function_handler_async(handler):
                res = await handler(
                    attempt=call.ctx.attempt,
                    event=call.event,
                    events=call.events,
                    run_id=call.ctx.run_id,
                    step=step_lib.Step(
                        client,
                        call.steps,
                        step_lib.StepIDCounter(),
                    ),
                )
            elif _is_function_handler_sync(handler):
                res = handler(
                    attempt=call.ctx.attempt,
                    event=call.event,
                    events=call.events,
                    run_id=call.ctx.run_id,
                    step=step_lib.StepSync(
                        client,
                        call.steps,
                        step_lib.StepIDCounter(),
                    ),
                )
            else:
                # Should be unreachable.
                return execution.CallError.from_error(
                    errors.UnknownError(
                        "unable to determine function handler type"
                    )
                )

            return res
        except step_lib.Interrupt as out:
            return [
                execution.CallResponse(
                    data=out.data,
                    display_name=out.display_name,
                    id=out.hashed_id,
                    name=out.name,
                    op=out.op,
                    opts=out.opts,
                )
            ]
        except Exception as err:
            # An error occurred with the user's code. Print the traceback to
            # help them debug.
            print(transforms.get_traceback(err))

            return execution.CallError.from_error(err)

    def call_sync(
        self,
        call: execution.Call,
        client: client_lib.Inngest,
        fn_id: str,
    ) -> (
        list[execution.CallResponse] | types.Serializable | execution.CallError
    ):
        try:
            handler: FunctionHandlerAsync | FunctionHandlerSync
            if self.id == fn_id:
                handler = self._handler
            elif self.on_failure_fn_id == fn_id:
                if self._opts.on_failure is None:
                    return execution.CallError.from_error(
                        errors.MissingFunction("on_failure not defined")
                    )
                handler = self._opts.on_failure
            else:
                return execution.CallError.from_error(
                    errors.MissingFunction("function ID mismatch")
                )

            if _is_function_handler_sync(handler):
                return handler(
                    attempt=call.ctx.attempt,
                    event=call.event,
                    events=call.events,
                    run_id=call.ctx.run_id,
                    step=step_lib.StepSync(
                        client,
                        call.steps,
                        step_lib.StepIDCounter(),
                    ),
                )
            return execution.CallError.from_error(
                errors.MismatchedSync(
                    "encountered async function in non-async context"
                )
            )
        except step_lib.Interrupt as out:
            return [
                execution.CallResponse(
                    data=out.data,
                    display_name=out.display_name,
                    id=out.hashed_id,
                    name=out.name,
                    op=out.op,
                    opts=out.opts,
                )
            ]
        except Exception as err:
            # An error occurred with the user's code. Print the traceback to
            # help them debug.
            print(transforms.get_traceback(err))

            return execution.CallError.from_error(err)

    def get_config(self, app_url: str) -> _Config:
        fn_id = self._opts.id

        name = fn_id
        if self._opts.name is not None:
            name = self._opts.name

        if self._opts.retries is not None:
            retries = function_config.Retries(attempts=self._opts.retries)
        else:
            retries = None

        main = function_config.FunctionConfig(
            batch_events=self._opts.batch_events,
            cancel=self._opts.cancel,
            debounce=self._opts.debounce,
            id=fn_id,
            name=name,
            rate_limit=self._opts.rate_limit,
            steps={
                const.ROOT_STEP_ID: function_config.Step(
                    id=const.ROOT_STEP_ID,
                    name=const.ROOT_STEP_ID,
                    retries=retries,
                    runtime=function_config.Runtime(
                        type="http",
                        url=f"{app_url}?fnId={fn_id}&stepId={const.ROOT_STEP_ID}",
                    ),
                ),
            },
            throttle=self._opts.throttle,
            triggers=[self._trigger],
        )

        on_failure = None
        if self.on_failure_fn_id is not None:
            on_failure = function_config.FunctionConfig(
                id=self.on_failure_fn_id,
                name=f"{name} (on_failure handler)",
                steps={
                    const.ROOT_STEP_ID: function_config.Step(
                        id=const.ROOT_STEP_ID,
                        name=const.ROOT_STEP_ID,
                        retries=function_config.Retries(attempts=0),
                        runtime=function_config.Runtime(
                            type="http",
                            url=f"{app_url}?fnId={self.on_failure_fn_id}&stepId={const.ROOT_STEP_ID}",
                        ),
                    )
                },
                triggers=[
                    function_config.TriggerEvent(
                        event=const.InternalEvents.FUNCTION_FAILED.value,
                        expression=f"event.data.function_id == '{self.id}'",
                    )
                ],
            )

        return _Config(main=main, on_failure=on_failure)

    def get_id(self) -> str:
        return self._opts.id



import json

import fastapi

from ._internal import client_lib, comm, const, execution, function, net


def serve(
    app: fastapi.FastAPI,
    client: client_lib.Inngest,
    functions: list[function.Function],
    *,
    base_url: str | None = None,
    signing_key: str | None = None,
) -> None:
    handler = comm.CommHandler(
        base_url=base_url or client.base_url,
        client=client,
        framework=const.Framework.FAST_API,
        functions=functions,
        logger=client.logger,
        signing_key=signing_key,
    )

    @app.get("/api/inngest")
    async def get_api_inngest(
        request: fastapi.Request,
    ) -> fastapi.Response:
        headers = net.normalize_headers(dict(request.headers.items()))
        is_from_dev_server = (
            headers.get(const.HeaderKey.SERVER_KIND.value)
            == const.ServerKind.DEV_SERVER.value
        )

        return _to_response(handler.inspect(is_from_dev_server))

    @app.post("/api/inngest")
    async def post_inngest_api(
        fnId: str,  # pylint: disable=invalid-name
        request: fastapi.Request,
    ) -> fastapi.Response:
        body = await request.body()
        headers = net.normalize_headers(dict(request.headers.items()))

        return _to_response(
            await handler.call_function(
                call=execution.Call.from_dict(json.loads(body)),
                fn_id=fnId,
                req_sig=net.RequestSignature(
                    body=body,
                    headers=headers,
                    is_production=client.is_production,
                ),
            )
        )

    @app.put("/api/inngest")
    async def put_inngest_api(request: fastapi.Request) -> fastapi.Response:
        headers = net.normalize_headers(dict(request.headers.items()))

        return _to_response(
            await handler.register(
                app_url=str(request.url),
                is_from_dev_server=(
                    headers.get(const.HeaderKey.SERVER_KIND.value)
                    == const.ServerKind.DEV_SERVER.value
                ),
            )
        )


def _to_response(comm_res: comm.CommResponse) -> fastapi.responses.JSONResponse:
    return fastapi.responses.JSONResponse(
        content=comm_res.body,
        headers=comm_res.headers,
        status_code=comm_res.status_code,
    )



from __future__ import annotations

import enum
import typing

from . import errors, event_lib, transforms, types


class Call(types.BaseModel):
    ctx: CallContext
    event: event_lib.Event
    events: list[event_lib.Event]
    steps: dict[str, object]


class CallContext(types.BaseModel):
    attempt: int
    run_id: str
    stack: CallStack


class CallStack(types.BaseModel):
    stack: list[str]


class CallError(types.BaseModel):
    is_internal: bool
    is_retriable: bool
    message: str
    name: str
    stack: str

    @classmethod
    def from_error(cls, err: Exception) -> CallError:
        return cls(
            is_internal=isinstance(err, errors.InternalError),
            is_retriable=isinstance(err, errors.NonRetriableError) is False,
            message=str(err),
            name=type(err).__name__,
            stack=transforms.get_traceback(err),
        )


class CallResponse(types.BaseModel):
    data: object
    display_name: str
    id: str
    name: str
    op: Opcode
    opts: dict[str, object] | None = None


def is_call_responses(
    value: object,
) -> typing.TypeGuard[list[CallResponse]]:
    if not isinstance(value, list):
        return False
    return all(isinstance(item, CallResponse) for item in value)


class Opcode(enum.Enum):
    SLEEP = "Sleep"
    STEP = "Step"
    WAIT_FOR_EVENT = "WaitForEvent"



import datetime
import hashlib
import json
import re
import traceback

from . import errors, result, types


def get_traceback(err: Exception) -> str:
    return "".join(
        traceback.format_exception(type(err), err, err.__traceback__)
    )


def hash_signing_key(key: str) -> str:
    return hashlib.sha256(
        bytearray.fromhex(remove_signing_key_prefix(key))
    ).hexdigest()


def hash_step_id(step_id: str) -> str:
    return hashlib.sha1(step_id.encode("utf-8")).hexdigest()


def dump_json(obj: object) -> result.Result[str, errors.UnserializableOutput]:
    try:
        return result.Ok(json.dumps(obj))
    except Exception as err:
        return result.Err(errors.UnserializableOutput(str(err)))


def remove_signing_key_prefix(key: str) -> str:
    prefix_match = re.match(r"^signkey-[\w]+-", key)
    prefix = ""
    if prefix_match:
        prefix = prefix_match.group(0)

    return key[len(prefix) :]


def prep_body(obj: types.T) -> types.T:
    """
    Prep body before sending to the Inngest server. This function will:
    - Remove items whose value is None.
    - Convert keys to camelCase.
    """

    if isinstance(obj, dict):
        return {
            to_camel_case(k): prep_body(v)
            for k, v in obj.items()
            if v is not None
        }  # type: ignore
    if isinstance(obj, list):
        return [prep_body(v) for v in obj if v is not None]  # type: ignore
    return obj


def to_camel_case(value: str) -> str:
    """
    Convert a string from snake_case to camelCase.
    """

    return "".join(
        word.title() if i else word for i, word in enumerate(value.split("_"))
    )


class _Duration:
    @classmethod
    def second(cls, count: int = 1) -> int:
        return count * 1000

    @classmethod
    def minute(cls, count: int = 1) -> int:
        return count * cls.second(60)

    @classmethod
    def hour(cls, count: int = 1) -> int:
        return count * cls.minute(60)

    @classmethod
    def day(cls, count: int = 1) -> int:
        return count * cls.hour(24)

    @classmethod
    def week(cls, count: int = 1) -> int:
        return count * cls.day(7)


def to_duration_str(
    ms: int | datetime.timedelta,
) -> result.Result[str, Exception]:
    if isinstance(ms, datetime.timedelta):
        ms = int(ms.total_seconds() * 1000)

    if ms < _Duration.second():
        return result.Err(
            errors.InvalidConfig("duration must be at least 1 second")
        )
    if ms < _Duration.minute():
        return result.Ok(f"{ms // _Duration.second()}s")
    if ms < _Duration.hour():
        return result.Ok(f"{ms // _Duration.minute()}m")
    if ms < _Duration.day():
        return result.Ok(f"{ms // _Duration.hour()}h")
    if ms < _Duration.week():
        return result.Ok(f"{ms // _Duration.day()}d")

    return result.Ok(f"{ms // _Duration.week()}w")


def to_iso_utc(value: datetime.datetime) -> str:
    return (
        value.astimezone(datetime.timezone.utc).strftime(
            "%Y-%m-%dT%H:%M:%S.%f"
        )[:-3]
        + "Z"
    )



from __future__ import annotations

import threading

from inngest._internal import execution, transforms, types


class StepBase:
    _memos: dict[str, object]
    _step_id_counter: StepIDCounter

    def _get_hashed_id(self, step_id: str) -> str:
        id_count = self._step_id_counter.increment(step_id)
        if id_count > 1:
            step_id = f"{step_id}:{id_count - 1}"
        return transforms.hash_step_id(step_id)

    def _get_memo(self, hashed_id: str) -> object:
        if hashed_id in self._memos:
            return self._memos[hashed_id]

        return types.EmptySentinel


class StepIDCounter:
    def __init__(self) -> None:
        self._counts: dict[str, int] = {}
        self._mutex = threading.Lock()

    def increment(self, hashed_id: str) -> int:
        with self._mutex:
            if hashed_id not in self._counts:
                self._counts[hashed_id] = 0

            self._counts[hashed_id] += 1
            return self._counts[hashed_id]


# Extend BaseException to avoid being caught by the user's code. Users can still
# catch it if they do a "bare except", but that's a known antipattern in the
# Python world.
class Interrupt(BaseException):
    def __init__(
        self,
        *,
        data: object = None,
        display_name: str,
        hashed_id: str,
        name: str,
        op: execution.Opcode,
        opts: dict[str, object] | None = None,
    ) -> None:
        self.data = data
        self.display_name = display_name
        self.hashed_id = hashed_id
        self.name = name
        self.op = op
        self.opts = opts


class WaitForEventOpts(types.BaseModel):
    if_exp: str | None
    timeout: str



from __future__ import annotations

import http
import logging
import os
import urllib.parse

import httpx

from inngest._internal import (
    client_lib,
    const,
    errors,
    execution,
    function,
    function_config,
    net,
    registration,
    result,
    transforms,
    types,
)


class CommResponse:
    _body: object | None = None

    def __init__(
        self,
        *,
        body: object | None = None,
        headers: dict[str, str],
        status_code: int = 200,
    ) -> None:
        self.headers = headers
        self.body = body
        self.status_code = status_code

    @property
    def body(self) -> object:
        return self._body or {}

    @body.setter
    def body(self, body: object) -> None:
        self._body = body

        if isinstance(body, (dict, list)):
            self.headers[
                const.HeaderKey.CONTENT_TYPE.value
            ] = "application/json"
        else:
            self.headers[const.HeaderKey.CONTENT_TYPE.value] = "text/plain"

    @classmethod
    def from_error(
        cls,
        err: Exception,
        framework: const.Framework,
    ) -> CommResponse:
        code = const.ErrorCode.UNKNOWN.value
        status_code = http.HTTPStatus.INTERNAL_SERVER_ERROR.value
        if isinstance(err, errors.InternalError):
            code = err.code
            status_code = err.status_code

        return cls(
            body={
                "code": code,
                "message": str(err),
            },
            headers=net.create_headers(framework=framework),
            status_code=status_code,
        )


class CommHandler:
    _base_url: str
    _client: client_lib.Inngest
    _fns: dict[str, function.Function]
    _framework: const.Framework
    _is_production: bool
    _logger: logging.Logger
    _signing_key: str | None

    def __init__(
        self,
        *,
        base_url: str | None = None,
        client: client_lib.Inngest,
        framework: const.Framework,
        functions: list[function.Function],
        logger: logging.Logger,
        signing_key: str | None = None,
    ) -> None:
        self._is_production = client.is_production
        self._logger = logger

        if not self._is_production:
            self._logger.info("Dev Server mode enabled")

        base_url = base_url or os.getenv(const.EnvKey.BASE_URL.value)
        if base_url is None:
            if not self._is_production:
                self._logger.info("Defaulting API origin to Dev Server")
                base_url = const.DEV_SERVER_ORIGIN
            else:
                base_url = const.DEFAULT_API_ORIGIN

        try:
            self._api_origin = net.parse_url(base_url)
        except Exception as err:
            raise errors.InvalidBaseURL() from err

        self._client = client
        self._fns = {fn.get_id(): fn for fn in functions}
        self._framework = framework

        if signing_key is None:
            if self._client.is_production:
                signing_key = os.getenv(const.EnvKey.SIGNING_KEY.value)
                if signing_key is None:
                    self._logger.error("missing signing key")
                    raise errors.MissingSigningKey()
        self._signing_key = signing_key

    def _build_registration_request(
        self,
        app_url: str,
    ) -> result.Result[httpx.Request, Exception]:
        registration_url = urllib.parse.urljoin(
            self._api_origin,
            "/fn/register",
        )

        match self.get_function_configs(app_url):
            case result.Ok(fn_configs):
                pass
            case result.Err(err):
                return result.Err(err)

        match registration.RegisterRequest(
            app_name=self._client.app_id,
            deploy_type=registration.DeployType.PING,
            framework=self._framework,
            functions=fn_configs,
            sdk=f"{const.LANGUAGE}:v{const.VERSION}",
            url=app_url,
            v="0.1",
        ).to_dict():
            case result.Ok(body):
                body = transforms.prep_body(body)
            case result.Err(err):
                return result.Err(err)

        headers = net.create_headers(framework=self._framework)
        if self._signing_key:
            headers[
                "Authorization"
            ] = f"Bearer {transforms.hash_signing_key(self._signing_key)}"

        return result.Ok(
            httpx.Client().build_request(
                "POST",
                registration_url,
                headers=headers,
                json=body,
                timeout=30,
            )
        )

    async def call_function(
        self,
        *,
        call: execution.Call,
        fn_id: str,
        req_sig: net.RequestSignature,
    ) -> CommResponse:
        """
        Handles a function call from the Executor.
        """

        validation_res = req_sig.validate(self._signing_key)
        if result.is_err(validation_res):
            return CommResponse.from_error(
                validation_res.err_value, self._framework
            )

        match self._get_function(fn_id):
            case result.Ok(fn):
                pass
            case result.Err(err):
                return CommResponse.from_error(err, self._framework)

        return self._create_response(await fn.call(call, self._client, fn_id))

    def call_function_sync(
        self,
        *,
        call: execution.Call,
        fn_id: str,
        req_sig: net.RequestSignature,
    ) -> CommResponse:
        """
        Handles a function call from the Executor.
        """

        validation_res = req_sig.validate(self._signing_key)
        if result.is_err(validation_res):
            err = validation_res.err_value
            extra = {}
            if isinstance(err, errors.InternalError):
                extra["code"] = err.code
            self._logger.error(err, extra=extra)
            return CommResponse.from_error(err, self._framework)

        match self._get_function(fn_id):
            case result.Ok(fn):
                pass
            case result.Err(err):
                extra = {}
                if isinstance(err, errors.InternalError):
                    extra["code"] = err.code
                self._logger.error(err, extra=extra)
                return CommResponse.from_error(err, self._framework)

        return self._create_response(fn.call_sync(call, self._client, fn_id))

    def _create_response(
        self,
        call_res: list[execution.CallResponse]
        | types.Serializable
        | execution.CallError,
    ) -> CommResponse:
        comm_res = CommResponse(
            headers={
                **net.create_headers(framework=self._framework),
                const.HeaderKey.SERVER_TIMING.value: "handler",
            }
        )

        if execution.is_call_responses(call_res):
            out: list[dict[str, object]] = []
            for item in call_res:
                match item.to_dict():
                    case result.Ok(d):
                        out.append(d)
                    case result.Err(err):
                        return CommResponse.from_error(err, self._framework)

            comm_res.body = transforms.prep_body(out)
            comm_res.status_code = 206
        elif isinstance(call_res, execution.CallError):
            match call_res.to_dict():
                case result.Ok(d):
                    body = transforms.prep_body(d)
                case result.Err(err):
                    return CommResponse.from_error(err, self._framework)

            self._logger.error(
                call_res.message,
                extra={"is_internal": call_res.is_internal},
            )
            comm_res.body = body
            comm_res.status_code = 500

            if call_res.is_retriable is False:
                comm_res.headers[const.HeaderKey.NO_RETRY.value] = "true"
        else:
            comm_res.body = call_res

        return comm_res

    def _get_function(
        self, fn_id: str
    ) -> result.Result[function.Function, Exception]:
        # Look for the function ID in the list of user functions, but also
        # look for it in the list of on_failure functions.
        for _fn in self._fns.values():
            if _fn.get_id() == fn_id:
                return result.Ok(_fn)
            if _fn.on_failure_fn_id == fn_id:
                return result.Ok(_fn)

        return result.Err(errors.MissingFunction(f"function {fn_id} not found"))

    def get_function_configs(
        self,
        app_url: str,
    ) -> result.Result[list[function_config.FunctionConfig], Exception]:
        configs: list[function_config.FunctionConfig] = []
        for fn in self._fns.values():
            config = fn.get_config(app_url)
            configs.append(config.main)

            if config.on_failure is not None:
                configs.append(config.on_failure)

        if len(configs) == 0:
            return result.Err(errors.InvalidConfig("no functions found"))
        return result.Ok(configs)

    def inspect(self, is_from_dev_server: bool) -> CommResponse:
        """
        Used by Dev Server to discover apps.
        """

        if is_from_dev_server and self._is_production:
            # Tell Dev Server to leave the app alone since it's in production
            # mode.
            return CommResponse(
                body={},
                headers={},
                status_code=403,
            )

        return CommResponse(
            body={},
            headers=net.create_headers(framework=self._framework),
            status_code=200,
        )

    def _parse_registration_response(
        self,
        server_res: httpx.Response,
    ) -> CommResponse:
        comm_res = CommResponse(
            headers=net.create_headers(framework=self._framework)
        )
        body: dict[str, object] = {}

        server_res_body: dict[str, object] | None = None
        try:
            raw_body: object = server_res.json()
            if isinstance(raw_body, dict):
                server_res_body = raw_body
        except Exception:
            pass

        if server_res.status_code < 400:
            if server_res_body:
                message = server_res_body.get("message")
                if isinstance(message, str):
                    body["message"] = message

                modified = server_res_body.get("modified")
                if isinstance(modified, bool):
                    body["modified"] = modified
        else:
            extra: dict[str, object] = {"status_code": server_res.status_code}

            if server_res_body:
                error = server_res_body.get("error")
                if isinstance(error, str):
                    body["message"] = error
                    extra["error"] = server_res_body.get("error")

            self._logger.error(
                "registration response failed",
                extra=extra,
            )

            comm_res.status_code = server_res.status_code

        comm_res.body = body
        return comm_res

    async def register(
        self,
        *,
        app_url: str,
        is_from_dev_server: bool,
    ) -> CommResponse:
        """
        Handles a registration call.
        """

        match self._validate_registration(is_from_dev_server):
            case result.Ok(_):
                pass
            case result.Err(err):
                print(err)
                self._logger.error(err)
                return CommResponse.from_error(err, self._framework)

        async with httpx.AsyncClient() as client:
            match self._build_registration_request(app_url):
                case result.Ok(req):
                    res = self._parse_registration_response(
                        await client.send(req)
                    )
                case result.Err(err):
                    print(err)
                    self._logger.error(err)
                    return CommResponse.from_error(err, self._framework)

        return res

    def register_sync(
        self,
        *,
        app_url: str,
        is_from_dev_server: bool,
    ) -> CommResponse:
        """
        Handles a registration call.
        """

        match self._validate_registration(is_from_dev_server):
            case result.Ok(_):
                pass
            case result.Err(err):
                return CommResponse.from_error(err, self._framework)

        with httpx.Client() as client:
            match self._build_registration_request(app_url):
                case result.Ok(req):
                    res = self._parse_registration_response(client.send(req))
                case result.Err(err):
                    return CommResponse.from_error(err, self._framework)

        return res

    def _validate_registration(
        self,
        is_from_dev_server: bool,
    ) -> result.Result[None, Exception]:
        if is_from_dev_server and self._is_production:
            return result.Err(
                errors.DevServerRegistrationNotAllowed(
                    "Dev Server registration not allowed in production mode"
                )
            )
        return result.Ok(None)
